from abc import ABC, abstractmethod
from .mixins.bash_command_mixin import BashCommandMixin
from .mixins.python_command_mixin import PythonCommandMixin
from .mixins.threaded_command_mixin import ThreadedCommandMixin


class ICommand(ABC):

    @abstractmethod
    def validate(self):
        pass

    @abstractmethod
    def execute(self):
        pass


class Command(ICommand):

    def __init__(self, command_config: dict):
        self.command_config = command_config
        self.mixins = {
            'bash': BashCommandMixin,
            'python': PythonCommandMixin,
            'threaded': ThreadedCommandMixin
        }
        command_type = self.command_config['type']

        if command_type not in self.mixins:
            raise ValueError(f"Unsupported command type '{command_type}'")

        mixin = self.mixins[command_type](self)
        self.mixin = mixin

    def validate(self):
        self.mixin.validate()

    def execute(self):
        if self.command_config.get("skip"):
            return
        self.mixin.execute()

    def create_command_instance(self, command_config):
        return Command(command_config)

    def execute_another_command(self, command_config):
        another_command = self.create_command_instance(command_config)
        another_command.validate()
        another_command.execute()



from abc import ABC, abstractmethod
from nanolab.loggers.sources.rpc_logger import RPCLogger


class LoggerBuilder(ABC):

    def __init__(self):
        self.timeout = None

    def set_timeout(self, timeout):
        self.timeout = timeout
        return self

    @abstractmethod
    async def build(self):
        pass


class RPCLoggerBuilder(LoggerBuilder):

    def __init__(self):
        super().__init__()
        self.rpc_url = None
        self.expected_blocks_count = None
        self.node_name = None

    def set_rpc_url(self, rpc_url):
        self.rpc_url = rpc_url
        return self

    def set_expected_blocks_count(self, expected_blocks_count):
        self.expected_blocks_count = expected_blocks_count
        return self

    def set_node_name(self, node_name):
        self.node_name = node_name
        return self

    async def build(self):
        if any(param is None for param in [self.rpc_url, self.expected_blocks_count, self.node_name, self.timeout]):
            raise Exception("Missing required parameters")
        logger = RPCLogger(self.node_name, self.rpc_url,
                           self.expected_blocks_count, self.timeout)
        await logger.async_init()
        return logger


# class WebSocketLoggerBuilder(LoggerBuilder):

#     def __init__(self):
#         super().__init__()
#         self.websocket_url = None
#         self.channel = None

#     def set_websocket_url(self, websocket_url):
#         self.websocket_url = websocket_url
#         return self

#     def set_channel(self, channel):
#         self.channel = channel
#         return self

#     def build(self):
#         if self.websocket_url is None or self.channel is None:
#             raise Exception("Missing required parameters")
#         # Assuming a hypothetical WebSocketLogger class
#         return WebSocketLogger(self.websocket_url, self.channel, self.timeout)



from abc import ABC, abstractmethod
from nanolab.publisher.block_asserts import BlockAsserts
from nanolab.publisher.block_generator import BlockGenerator
from nanolab.publisher.confirmation_stats import ConfirmationStatsManager
from nanolab.src.utils import get_config_parser, print_dot
from nanolab.publisher.event_bus import EventBus



class ITestCase(ABC):

    @abstractmethod
    def __init__(self, config: dict):
        pass

    @abstractmethod
    async def run(self):
        pass


class TestCaseFactory:

    @staticmethod
    def create(config: dict) -> ITestCase:
        # The config should include a 'type' field that identifies the test case type.
        if config['type'] == 'change_wait_conf':
            return ChangeBlockWaitConfirmation(config)
        if config['type'] == 'send_wait_conf':
            return SendBlockWaitConfirmation(config)
        # Add more elif branches here for other test case types.
        else:
            raise ValueError(f"Invalid test case type: {config['type']}")


class ChangeBlockWaitConfirmation(ITestCase):

    def __init__(self, config: dict):
        self.event_bus = EventBus()
        conf_p = get_config_parser()

        self.source_seed = config["source_seed"]
        self.start_index = config["source_start_index"]
        self.block_count = config["block_count"]
        self.is_independent = config.get("is_independent", True)
        self.timeout_s = config.get("timeout_s", 10)
        self.node_name = config.get("node_name", conf_p.get_nodes_name()[:-1])

        self.ba_l = BlockAsserts(self.event_bus, self.node_name)
        self.bg_l = BlockGenerator(self.node_name)
        self.stats_manager = ConfirmationStatsManager(self.timeout_s)
        self.event_bus.subscribe('block_confirmed',
                                 self.stats_manager.on_block_confirmed)

    @print_dot
    async def _generate_and_confirm_block(self, index):
        self.bg_l.set_broadcast_blocks(True)
        representative = self.bg_l.get_random_account()
        res = await self.bg_l.blockgen_single_change(source_seed=self.source_seed,
                                                     source_index=index,
                                                     rep=representative)
        try:
            await self.ba_l.assert_single_block_confirmed_wait(
                res["hash"], self.timeout_s, 0.05)
        except AssertionError as ex:
            print("DEBUG", ex)

    async def run(self):
        await self.stats_manager.initialize(self.node_name)
        for counter in range(self.start_index,
                             self.start_index + self.block_count):
            seed_index = counter if self.is_independent else self.start_index
            await self._generate_and_confirm_block(seed_index)

        await self.stats_manager.print_stats()  # no longer pass conf_lst
        return self.block_count


class SendBlockWaitConfirmation(ITestCase):

    def __init__(self, config: dict):
        self.event_bus = EventBus()
        conf_p = get_config_parser()

        self.source_seed = config["source_seed"]
        self.destination_seed = config["destination_seed"]
        self.start_index = config["source_start_index"]
        self.send_amount_raw = config["send_amount_raw"]
        self.block_count = config["block_count"]
        self.is_independent = config.get("is_independent", True)
        self.timeout_s = config.get("timeout_s", 10)
        self.node_name = config.get("node_name", conf_p.get_nodes_name()[:-1])

        self.ba_l = BlockAsserts(self.event_bus, self.node_name)
        self.bg_l = BlockGenerator(self.node_name)
        self.stats_manager = ConfirmationStatsManager(self.timeout_s)
        self.event_bus.subscribe('block_confirmed',
                                 self.stats_manager.on_block_confirmed)

    @print_dot
    async def _generate_and_confirm_block(self, index):
        self.bg_l.set_broadcast_blocks(True)
        representative = self.bg_l.get_random_account()
        res = await self.bg_l.create_send_and_open_block(self.send_amount_raw,
                                                         self.source_seed, index,
                                                         self.destination_seed,
                                                         index, representative)
        try:
            await self.ba_l.assert_blocks_confirmed_wait(
                [block["hash"] for block in res], self.timeout_s, 0.05)
        except AssertionError as ex:
            print("DEBUG", ex)

    async def run(self):
        await self.stats_manager.initialize(self.node_name)
        for counter in range(self.start_index,
                             self.start_index + self.block_count):
            seed_index = counter if self.is_independent else self.start_index
            await self._generate_and_confirm_block(seed_index)

        await self.stats_manager.print_stats()  # no longer pass conf_lst
        return self.block_count



from nanolab.loggers.contracts import ISink, LogData
import datetime
import time


class ConsoleSink(ISink):
    """A storage that writes logs to the console."""
    percent_cemented = 0

    def __init__(self, **kwargs):
        self.pass_above = kwargs.get("pass_above")
        self.pass_at = kwargs.get("pass_at", 100)

    def store_logs(self, logs: LogData):
        """Store logs by printing them to the console."""
        node_name = logs.node_name
        node_version = logs.node_version
        timestamp = logs.timestamp
        elapsed_time = logs.elapsed_time
        check_count = logs.check_count
        cemented_count = logs.cemented_count
        bps = logs.bps
        cps = logs.cps
        percent_cemented = logs.percent_cemented
        percent_checked = logs.percent_checked
        bps_avg = logs.bps_avg or 0
        cps_avg = logs.cps_avg or 0

        self.percent_cemented = percent_cemented

        print(
            f"{timestamp:<20} {elapsed_time:>4} sec | {node_name[:16]:<16} | {node_version:<10} | \
{cemented_count:>7}/{check_count:>7} @ CPS: {cps:>7} (avg {cps_avg:>7.2f}) ({percent_cemented:>6.2f}%) | \
BPS: {bps:>7} (avg {bps_avg:>7.2f}) ({percent_checked:>6.2f}%)")

    def end(self):
        if self.pass_above:
            status = "PASS" if self.percent_cemented >= self.pass_above else "FAIL"
        else:
            status = "PASS" if self.percent_cemented == self.pass_at else "FAIL"
        print(status)



import shutil
from pathlib import Path
from os import environ
from nanomock.modules.nl_parse_config import ConfigReadWrite
from nanolab.src.utils import download_data
from nanolab.src.github import GitHubAPI


class ConfigPathHandler:

    def __init__(self, testcase_alias: str):
        self.conf_rw = ConfigReadWrite()

        self.is_local_path = False
        self.local_config_path = None
        self.base_path = Path(environ.get("NL_PATH") or str(Path.cwd()))
        self.testcase_name = self._set_testcase_name(testcase_alias.strip())
        self.resources_dir = environ.get("NL_RES_DIR", "testcases")
        self.resources_path = self.base_path / self.resources_dir
        self.downloads_path = self.base_path / self.resources_dir / "downloads"
        self.config_copy_file_path = self.resources_path / \
            f"{self.testcase_name}_config.json"
        self.resolved_config_file_path = self.resources_path / "resolved_config.json"
        self.snippets_path = self.base_path / "snippets"

        self.resources_path.mkdir(parents=True, exist_ok=True)
        self.downloads_path.mkdir(parents=True, exist_ok=True)

    def _set_testcase_name(self, testcase_alias):
        # for ease of use we allow alias to be either local path to disk, or downloadable alias
        if testcase_alias.endswith(".json"):
            self.local_config_path = testcase_alias
            config = self.conf_rw.read_json(testcase_alias)
            testcase_name = config.get("testcase", "unknown_testcase")
            self.is_local_path = True
        else:
            testcase_name = testcase_alias

        environ["LAB_TESTCASE"] = testcase_name
        return testcase_name

    def get_testcase_name(self):
        return self.testcase_name

    def get_resources_dir(self):
        return self.resources_dir

    def get_resources_path(self):
        return self.resources_path

    def get_snippets_path(self):
        return self.snippets_path

    def get_resolved_config_path(self):
        return self.resolved_config_file_path

    def _copy_path(self, path: str) -> str:
        filename = Path(path).name
        destination = self.resources_path / filename

        copy_files = environ.get("NL_COPY_FILES", "True")
        if copy_files.lower() == "true" and not destination.exists():
            shutil.copy(path, destination)
        return destination


class ConfigResourceHandler():

    def __init__(self, github_api: GitHubAPI,
                 config_path_handler: ConfigPathHandler):

        self.conf_rw = ConfigReadWrite()
        self.config_path = config_path_handler.get_resources_path()
        self.testcase_name = config_path_handler.testcase_name
        self.github_api = github_api
        self.is_local_path = config_path_handler.is_local_path
        self.local_config_path = config_path_handler.local_config_path

    def download_and_replace_resources(self, config_path: Path,
                                       destination_path: Path):

        config = self.conf_rw.read_json(config_path)
        global_resources = config.get("global", {})

        for key, value in global_resources.items():
            value_string = str(value)
            if value_string.startswith("http://") or value_string.startswith("https://"):
                # Download the remote file
                downloaded_path = self._download_url(value, destination_path)
                # Replace the URL with the downloaded file path
                global_resources[key] = str(downloaded_path)
            elif environ.get("NL_COPY_FILES") and environ.get(
                    "NL_COPY_FILES").lower() == "true":
                # Copy the local file to the destination
                copied_path = self._copy_path(
                    value, destination_path / Path(value).name)
                # Replace the local path with the copied file path
                global_resources[key] = str(copied_path)

        return config

    def copy_config_file(self, destination: Path):
        if self.is_local_path:
            return self._copy_path(self.local_config_path,
                                   destination,
                                   force_copy=True)
        else:
            return self._download_config_file(destination)

    def _copy_path(self,
                   source_path: str,
                   destination: Path,
                   force_copy=None) -> str:
        source = Path(source_path)

        copy_files = force_copy or environ.get("NL_COPY_FILES", "False")
        if str(copy_files).lower() == "true" or not destination.exists():
            shutil.copy(source, destination)
        return destination

    def _download_url(self, url: str, resource_path: Path) -> Path:
        return download_data(url.strip(), Path(url).name, resource_path)

    def _download_config_file(self, destination: Path) -> Path:
        config_data = self.github_api.get_file_content(
            f"{self.testcase_name}.json")
        return download_data(config_data, destination.name, destination.parent)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional


class ILogger(ABC):

    @abstractmethod
    async def fetch_logs(self):
        pass


class ISink(ABC):

    @abstractmethod
    def store_logs(self, logs: dict):
        pass

    @abstractmethod
    def end(self):
        pass


@dataclass
class LogData:
    node_name: str
    node_version: str
    elapsed_time: int
    check_count: int
    cemented_count: int
    cps: int
    bps: int
    timestamp: str
    percent_cemented: float
    percent_checked: float
    bps_avg: Optional[float] = None
    cps_avg: Optional[float] = None


import threading
from nanolab import pycmd
from .base_mixin import CommandMixinBase
from collections import defaultdict
import heapq
import time
import copy


class ThreadedCommandMixin(CommandMixinBase):

    def validate(self):
        for command_config in self.command_instance.command_config["commands"]:
            command = self.command_instance.create_command_instance(
                command_config)
            command.validate()

    def execute(self):
        normalized_commands = self._normalize_delays(
            self.command_instance.command_config["commands"])
        schedule = self._build_schedule(normalized_commands)
        self._execute_commands(schedule)

    def _execute_command_sequence(self, commands):
        for command_config in commands:
            self.command_instance.execute_another_command(command_config)

    def _normalize_delays(self, commands):
        min_delay = min(cmd.get("delay", 0) for cmd in commands)
        return [{**cmd, "delay": cmd.get("delay", 0) + abs(min_delay)} for cmd in commands]

    def _build_schedule(self, commands):
        schedule = []
        command_groups = defaultdict(list)

        def _add_to_schedule(delay, target, args):
            run_at = time.time() + delay
            time.sleep(0.01)
            heapq.heappush(schedule, (run_at, target, args))

        def _append_replicas(command_config):
            # Extract and remove the replication count from the command_config
            replica_count = int(command_config.pop('replicas', 1))
            for _ in range(replica_count):
                # Make a deep copy to ensure each command is a separate object
                replicated_command = copy.deepcopy(command_config)
                delay = replicated_command.get("delay", 0)
                _add_to_schedule(
                    delay, self.command_instance.execute_another_command, (replicated_command,))

        for command_config in commands:
            group = command_config.get("group")
            if group:
                # If the command is part of a group, append it to the respective group list
                command_groups[group].append(command_config)
            else:
                # Check if this command should be independently replicated
                if "replicas" in command_config:
                    _append_replicas(command_config)
                else:
                    delay = command_config.get("delay", 0)
                    _add_to_schedule(
                        delay, self.command_instance.execute_another_command, (command_config,))

        for group, command_configs in command_groups.items():
            delay = command_configs[0].get("delay", 0)
            _add_to_schedule(
                delay, self._execute_command_sequence, (command_configs,))

        return schedule

    def _execute_commands(self, schedule):
        threads = []
        while schedule:
            run_at, target, args = heapq.heappop(schedule)
            delay = max(0, run_at - time.time())
            time.sleep(delay)
            thread = threading.Thread(target=target, args=args)
            thread.start()
            threads.append(thread)

        for thread in threads:
            thread.join()



from nanolab.loggers.contracts import ILogger, LogData
from nanomock.modules.nl_rpc import NanoRpc
import time
from datetime import datetime
import asyncio
from typing import AsyncIterator
import logging


class RPCLogger(ILogger):
    """A logger that fetches logs from an RPC endpoint."""
    FETCH_INTERVAL_DELAY = 0.1  # seconds

    def __init__(self, node_name: str, rpc_url: str, expected_blocks_count: int, timeout: int):
        self.node_name = node_name
        self.rpc_url = rpc_url
        self.expected_blocks_count = int(expected_blocks_count)
        self.timeout = int(timeout)
        self.nanorpc = NanoRpc(self.rpc_url)
        # These will be initialized in async_init
        self.count_start = None
        self.cemented_start = None
        self.node_version = None
        self.end_block_count = None
        self.previous_count = 0
        self.previous_cemented = 0
        self.previous_elapsed_time = 0

    async def async_init(self, count_start=None, cemented_start=None):
        if count_start is None or cemented_start is None:
            self.count_start, self.cemented_start = await self._get_block_count()
        else:
            self.count_start = count_start
            self.cemented_start = cemented_start
        node_version = await self.nanorpc.version()
        self.node_version = f'{node_version["node_vendor"]} {node_version["build_info"][0:7]}' if node_version else "???"
        self.end_block_count = self.count_start + self.expected_blocks_count

    async def _get_block_count(self):
        block_count = await self.nanorpc.block_count()
        return int(block_count["count"]), int(block_count["cemented"])

    def is_fully_synced(self, cemented):
        cemented_diff = cemented - self.cemented_start
        is_synced = cemented_diff >= self.expected_blocks_count
        return is_synced

    async def fetch_logs(self) -> AsyncIterator[LogData]:
        """Yield logs at intervals."""
        start_time = time.time()
        timeout_start = time.time()
        while True:
            count, cemented = await self._get_block_count()
            is_synced = self.is_fully_synced(cemented)
            percent_cemented = ((cemented - self.cemented_start) /
                                self.expected_blocks_count) * 100
            percent_checked = (
                (count - self.count_start) / self.expected_blocks_count) * 100

            # Loggers are cerated before the publishing starts.
            # Elapsed is started when the first blcok is received by the node
            if percent_checked == 0 and percent_cemented == 0:
                start_time = time.time()
            elapsed_time = int(time.time() - start_time)

            cps = (cemented - self.previous_cemented) / max(
                1, (elapsed_time -
                    self.previous_elapsed_time)) if percent_checked > 0 else 0
            bps = (count - self.previous_count) / max(
                1, (elapsed_time -
                    self.previous_elapsed_time)) if percent_checked > 0 else 0

            bps_avg = (
                count - self.count_start
            ) / elapsed_time if elapsed_time > 0 and percent_checked < 100 else None
            cps_avg = (cemented - self.cemented_start
                       ) / elapsed_time if elapsed_time > 0 else None

            yield LogData(
                node_name=self.node_name,
                node_version=self.node_version,
                elapsed_time=elapsed_time,
                check_count=count,
                cemented_count=cemented,
                cps=cps,
                bps=bps,
                timestamp=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                percent_cemented=percent_cemented,
                percent_checked=percent_checked,
                bps_avg=bps_avg,
                cps_avg=cps_avg)
            if is_synced or time.time() - timeout_start > self.timeout:
                break

            self.previous_count = count
            self.previous_cemented = cemented
            self.previous_elapsed_time = elapsed_time
            await asyncio.sleep(self.FETCH_INTERVAL_DELAY)


# class RPCLogger(ILogger):

#     def __init__(
#         self,
#         rpc_url: str,
#         expected_blocks_count: int,
#         timeout: int,
#     ):
#         self.nanorpc = NanoRpc(rpc_url)
#         self.expected_blocks_count = expected_blocks_count
#         self.timeout = timeout
#         self.count_start, self.cemented_start = self._get_block_count()

#     def _get_block_count(self):
#         block_count = self.nanorpc.block_count()
#         return int(block_count["count"]), int(block_count["cemented"])

#     def is_fully_synced(self):
#         count, cemented = self.nanorpc.block_count()
#         cemented_diff = cemented - self.cemented_start
#         is_synced = cemented_diff == self.expected_blocks_count
#         return is_synced, count, cemented

#     async def fetch_logs(self):
#         previous_count = 0
#         previous_cemented = 0
#         start_time = time.time()

#         while True:
#             is_synced, check_count, cemented_count = self.is_fully_synced()
#             elapsed_time = int(time.time() - start_time)

#             logs = {
#                 'elapsed_time': elapsed_time,
#                 'check_count': check_count,
#                 'cemented_count': cemented_count,
#                 'previous_count': previous_count,
#                 'previous_cemented': previous_cemented,
#             }

#             yield logs

#             if is_synced or elapsed_time > self.timeout:
#                 break

#             previous_count = check_count
#             previous_cemented = cemented_count
#             await asyncio.sleep(1)



#!/bin/env python3

import binascii

from nanolab.xnomin.peers import message_header, message_type, message_type_enum


class telemetry_req:

    def __init__(self, ctx: dict):
        self.header = message_header(
            ctx['net_id'], [21, 21, 20],
            message_type(message_type_enum.telemetry_req), 0)

    def serialise(self) -> bytes:
        return self.header.serialise_header()


class telemetry_ack:

    def __init__(self, hdr: message_header, signature: bytes, node_id: bytes,
                 block_count: int, cemented_count: int, unchecked_count: int,
                 account_count: int, bandwidth_cap: int, peer_count: int,
                 protocol_ver: int, uptime: int, genesis_hash: bytes,
                 major_ver: int, minor_ver: int, patch_ver: int,
                 pre_release_ver: int, maker_ver: int, timestamp: int,
                 active_difficulty: int):
        self.hdr = hdr
        self.sig_verified = False
        self.sig = signature
        self.node_id = node_id
        self.block_count = block_count
        self.cemented_count = cemented_count
        self.unchecked_count = unchecked_count
        self.account_count = account_count
        self.bandwidth_cap = bandwidth_cap
        self.peer_count = peer_count
        self.protocol_ver = protocol_ver
        self.uptime = uptime
        self.genesis_hash = genesis_hash
        self.major_ver = major_ver
        self.minor_ver = minor_ver
        self.patch_ver = patch_ver
        self.pre_release_ver = pre_release_ver
        self.maker_ver = maker_ver
        self.timestamp = timestamp
        self.active_difficulty = active_difficulty

    @classmethod
    def from_json(self, json_tel: dict):
        return telemetry_ack(
            message_header.from_json(json_tel['hdr']),
            binascii.unhexlify(json_tel['sig']),
            binascii.unhexlify(json_tel['node_id']), json_tel['block_count'],
            json_tel['cemented_count'], json_tel['unchecked_count'],
            json_tel['account_count'], json_tel['bandwidth_cap'],
            json_tel['peer_count'], json_tel['protocol_ver'],
            json_tel['uptime'], binascii.unhexlify(json_tel['genesis_hash']),
            json_tel['major_ver'], json_tel['minor_ver'],
            json_tel['patch_ver'], json_tel['pre_release_ver'],
            json_tel['maker_ver'], json_tel['timestamp'],
            json_tel['active_difficulty'])


from nanomock.modules.nl_parse_config import ConfigReadWrite
from pathlib import Path
import logging

logger = logging.getLogger(__name__)


class SnippetManager:

    def __init__(self, snippet_folder: Path):
        self.conf_rw = ConfigReadWrite()
        self.snippet_folder = snippet_folder
        self.snippets = {}
        self._load_snippets()

    def _load_snippets(self):
        snippet_files = list(self.snippet_folder.glob("*.json"))

        for snippet_file in snippet_files:
            try:
                snippets = self.conf_rw.read_json(snippet_file)
                for snippet_key, snippet_value in snippets.items():
                    self.snippets[snippet_key] = snippet_value
            except Exception as exc:
                logger.info(f"Failed to load snippets from {snippet_file}")

    def get_snippets(self):
        return self.snippets

    def get_snippet_by_key(self, snippet_by_key):
        return self.snippets[snippet_by_key]



from nanolab.loggers.builders import RPCLoggerBuilder
from typing import Dict, Any


class LoggerFactory:

    @staticmethod
    async def create_logger(logger_type: str, config: Dict[str, Any]):
        if logger_type == "rpc":
            builder = RPCLoggerBuilder()
        elif logger_type == "websocket":
            pass
            # builder = WebsocketLoggerBuilder()
        else:
            raise ValueError(f"Unknown logger type: {logger_type}")

        for key, value in config.items():
            if hasattr(builder, f'set_{key}'):
                getattr(builder, f'set_{key}')(value)

        return await builder.build()



from sqlalchemy import create_engine, Column, Integer, String, Float, ForeignKey
from sqlalchemy.orm import sessionmaker, relationship, declarative_base
from nanolab.loggers.contracts import ISink, LogData
from nanolab.decorators import print_dot
from os import environ

Base = declarative_base()


class SqlLog(Base):
    __tablename__ = "nanolab_logs"

    id = Column(Integer, primary_key=True, autoincrement=True)
    testcase_id = Column(Integer, ForeignKey('nanolab_testcases.id'))
    timestamp = Column(String)
    elapsed_time = Column(Integer)
    check_count = Column(Integer)
    cemented_count = Column(Integer)
    percent_cemented = Column(Float)
    percent_checked = Column(Float)
    cps_avg = Column(Float)


class SqlTestCase(Base):
    __tablename__ = "nanolab_testcases"

    id = Column(Integer, primary_key=True)
    testcase_name = Column(String)
    start_date = Column(String)
    end_date = Column(String)
    status = Column(String)

    logs = relationship("SqlLog", backref="testcase")


class SqlSink(ISink):

    def __init__(self, **kwargs):
        self.db_uri = kwargs['db_uri']
        self.milestones = kwargs['milestones']
        self.testcase_name = kwargs.get(
            'testcase_name', environ.get("LAB_TESTCASE", "UNDEFINED"))
        self.start_date = None
        self.end_date = None
        self.status = 'Running'

        self.engine = create_engine(self.db_uri)
        Base.metadata.create_all(self.engine)
        self.Session = sessionmaker(bind=self.engine)

        self.testcase_id = self.store_testcase()

    def store_testcase(self):
        session = self.Session()
        new_testcase = SqlTestCase(testcase_name=self.testcase_name,
                                   start_date=self.start_date,
                                   end_date=self.end_date,
                                   status=self.status)
        session.add(new_testcase)
        session.commit()
        testcase_id = new_testcase.id
        session.close()
        return testcase_id

    def update_testcase(self, status):
        session = self.Session()
        testcase = session.query(SqlTestCase).filter_by(
            id=self.testcase_id).first()
        testcase.end_date = self.end_date
        testcase.status = status
        session.commit()
        session.close()

    @print_dot
    def store_logs(self, logs: LogData):
        if self.milestones and logs.percent_cemented >= min(self.milestones):
            session = self.Session()
            new_log = SqlLog(testcase_id=self.testcase_id,
                             timestamp=logs.timestamp,
                             elapsed_time=logs.elapsed_time,
                             check_count=logs.check_count,
                             cemented_count=logs.cemented_count,
                             percent_cemented=logs.percent_cemented,
                             percent_checked=logs.percent_checked,
                             cps_avg=logs.cps_avg)
            session.add(new_log)
            session.commit()
            session.close()
            self.milestones = {
                milestone
                for milestone in self.milestones
                if milestone > logs.percent_cemented
            }
            if logs.percent_cemented == 100:
                self.status = "PASS"

    def end(self):
        status = "FAIL" if self.status == "RUNNING" else self.status
        self.update_testcase(status)

    def get_testcase(self, id):
        with self.Session() as session:
            return session.query(SqlTestCase).filter_by(id=id).first()

    def get_log(self, testcase_id):
        with self.Session() as session:
            return session.query(SqlLog).filter_by(
                testcase_id=testcase_id).first()



import subprocess
from .base_mixin import CommandMixinBase


class BashCommandMixin(CommandMixinBase):

    def validate(self):
        pass
        # Validation logic for BashCommand

    def execute(self):
        command_config = self.command_instance.command_config
        variables = self._get_completed_variables()
        command = self.safe_format(command_config['command'], **variables)
        background = command_config.get('background', False)

        if background:
            self.execute_background(command, command_config.get('pid'))
        else:
            self.execute_foreground(command)

    def execute_foreground(self, command):
        process = subprocess.Popen(
            command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

        # Stream the output
        while True:
            output = process.stdout.readline()
            if output == '' and process.poll() is not None:
                break
            if output:
                print(output.strip())

        # Check for any errors
        error_output = process.stderr.read()
        if error_output:
            print(
                f"Error executing command: {command}\nError: {error_output.strip()}")

    def execute_background(self, command, pid_file=None):
        process = subprocess.Popen(
            command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if pid_file:
            self.write_pid_to_file(process.pid, pid_file)
        print(
            f"Command '{command}' is running in the background with PID {process.pid}")

    def write_pid_to_file(self, pid, filename):
        with open(filename, 'w') as file:
            file.write(str(pid))

    def handle_command_output(self, command, process):
        if process.returncode != 0:
            print(
                f"Error executing command: {command}\nError: {process.stderr.strip()}")
        else:
            response = process.stdout.strip() or process.stderr.strip() or command
            print(response)

    def safe_format(self, template, **kwargs):
        class SafeDict(dict):
            def __missing__(self, key):
                return ''
        return template.format_map(SafeDict(**kwargs))



# nanolab/command/argparse_commands.py

from os import environ, listdir
from .github import GitHubAPI
import argparse
from nanolab.src.config_handler import ConfigPathHandler, ConfigResourceHandler
from nanomock.modules.nl_parse_config import ConfigReadWrite
from nanolab.src.utils import extract_packaged_data_to_disk
from nanolab.src.config_loader import ConfigLoader, ConfigValidator, ConfigCommandExecutor
from nanolab.src.snippet_manager import SnippetManager


def parse_args():
    parser = argparse.ArgumentParser(
        description="Load configuration and snippets files")

    subparsers = parser.add_subparsers(dest="command")

    # Run command
    run_parser = subparsers.add_parser("run", help="Run a testcase")
    run_parser.add_argument('-t', '--testcase', type=str, help='Testcase name')
    run_parser.add_argument('-i', '--image', nargs='+', help='Docker image(s)')
    run_parser.add_argument('--gh-user',
                            type=str,
                            default='gr0vity-dev',
                            help='GitHub user (default: gr0vity-dev)')
    run_parser.add_argument('--gh-repo', type=str, default='nanolab-configs',
                            help='GitHub repository (default: nanolab-configs)')
    run_parser.add_argument('--gh-path', type=str, default='default',
                            help='Path to testcases within the repository (default: default)')
    run_parser.add_argument('--gh-ref',  type=str,
                             help="Set github ref (equivalent to the branch)'")

    # List command
    list_parser = subparsers.add_parser("list", help="List testcases")
    list_parser.add_argument(
        '--gh-user', type=str, default='gr0vity-dev', help='GitHub user (default: gr0vity-dev)')
    list_parser.add_argument('--gh-repo', type=str, default='nanolab-configs',
                             help='GitHub repository (default: nanolab-configs)')
    list_parser.add_argument('--gh-path', type=str, default='default',
                             help='Path to testcases within the repository (default: default)')
    list_parser.add_argument('--gh-ref',  type=str,
                             help="Set github ref (equivalent to the branch)'")
    list_parser.add_argument('--local', action='store_true',
                             help="List testcases from local directory './testcases/'")
    list_parser.add_argument('--paths', action='store_true',
                             help="List all available paths'")
    list_parser.add_argument('--refs', action='store_true',
                             help="List all available branches'")
    

    return parser.parse_args()


class ArgParseHandler:

    def __init__(self, args=None):
        self.args = args or parse_args()
        self.conf_rw = ConfigReadWrite()
        self.github_api = GitHubAPI(self.args.gh_user, self.args.gh_repo,
                                    self.args.gh_path,  self.args.gh_ref)

    def run(self):
        path_handler = ConfigPathHandler(self.args.testcase)
        resource_handler = ConfigResourceHandler(self.github_api, path_handler)

        #
        extract_packaged_data_to_disk(path_handler.get_snippets_path())

        # copy config to disk
        config_path = resource_handler.copy_config_file(
            path_handler.config_copy_file_path)

        # resolve and downlaod resources defiend in the config
        resolved_config = resource_handler.download_and_replace_resources(
            config_path, path_handler.downloads_path)
        # replace docker_tags with commandline
        if self.args.image:
            resolved_config["docker_tags"] = self.args.image

        resolved_path = path_handler.get_resolved_config_path()
        snippet_path = path_handler.get_snippets_path()

        snippet_manager = SnippetManager(snippet_path)
        config_loader = ConfigLoader(snippet_manager)
        final_config = config_loader.load_config(resolved_config)
        self.conf_rw.write_json(resolved_path, final_config)
        ConfigValidator.validate_config(final_config)
        ConfigCommandExecutor.execute_commands(final_config)

    def list_testcases(self):
        if self.args.local:
            # Create a ConfigPathHandler instance
            # Placeholder since we don't need a specific testcase_alias for listing
            path_handler = ConfigPathHandler("placeholder")

            # Get the base path for test cases
            base_path = path_handler.get_resources_dir()

            # If --local flag is set, list test cases from the local directory
            for file in listdir(base_path):
                if file.endswith('_config.json') and file != 'resolved_config.json':
                    # Prepend the path and remove the '.json' extension before printing
                    print(f"{base_path}/{file}")
        elif self.args.paths:
            # Otherwise, list test cases from the GitHub repository
            branches = self.github_api.get_paths()
            if branches:
                for branch in branches:
                    print(branch)
        elif self.args.refs:
            # Otherwise, list test cases from the GitHub repository
            branches = self.github_api.get_branches()
            if branches:
                for branch in branches:
                    print(branch)     
        else:
            # Otherwise, list test cases from the GitHub repository
            files = self.github_api.get_files()
            if files:
                for file in files:
                    if file["name"].endswith(".json"):
                        print(file["name"][:-5])

    def get_args(self):
        return self.args



#!./venv_py/bin/python

import logging
from nanolab.src.argparse_commands import ArgParseHandler

logger = logging.getLogger(__name__)


def main():
    arg_parse = ArgParseHandler()
    args = arg_parse.get_args()
    if args.command == "run":
        arg_parse.run()

    elif args.command == "list":
        arg_parse.list_testcases()

    else:
        print("Invalid command. Use 'nanolab run' or 'nanolab list'")


if __name__ == "__main__":
    main()



from nanomock.modules.nl_rpc import NanoRpc, NanoLibTools
from nanolab.src.utils import get_config_parser
from abc import ABC, abstractmethod


class IBlockGenerator(ABC):

    @abstractmethod
    def __init__(self, node_name=None):
        pass

    @abstractmethod
    def get_random_account(self):
        pass

    @abstractmethod
    async def blockgen_single_change(self,
                                     source_seed=None,
                                     source_index=None,
                                     source_private_key=None,
                                     rep=None):
        pass

    @abstractmethod
    def set_broadcast_blocks(self, broadcast):
        pass


class BlockGenerator(IBlockGenerator):

    def __init__(self, node_name=None):
        conf_p = get_config_parser()
        node_name = node_name if node_name else conf_p.get_nodes_name()[:-1]
        self.nano_rpc_default = NanoRpc(conf_p.get_node_rpc(node_name))
        self.broadcast = False
        self.single_account_open_counter = 0
        self.nano_lib = NanoLibTools()

    def get_random_account(self):
        random_seed = self.nano_rpc_default.generate_seed()
        random_account = self.nano_lib.nanolib_account_data(seed=random_seed)
        return random_account["account"]

    async def blockgen_single_change(self,
                                     source_seed=None,
                                     source_index=None,
                                     source_private_key=None,
                                     rep=None):
        rep = rep if rep else self.get_random_account()
        if source_private_key:
            res = await self.nano_rpc_default.create_change_block_pkey(
                source_private_key, rep, broadcast=self.broadcast)
            return res

        if source_seed and source_index >= 0:
            return await self.nano_rpc_default.create_change_block(
                source_seed, source_index, rep, broadcast=self.broadcast)

        raise ValueError(
            "Either source_private_key or source_seed and source_index must not be None"
        )

    def set_broadcast_blocks(self, broadcast):
        self.broadcast = broadcast

    async def create_send_and_open_block(self, send_amount_raw, source_seed,
                                         source_index, destination_seed,
                                         destination_index, representative):
        # destination = self.nano_rpc_default.generate_account(destination_seed, destination_index)
        return await self.blockgen_single_account_opener(
            representative=representative,
            source_seed=source_seed,
            source_index=source_index,
            destination_seed=destination_seed,
            destination_index=destination_index,
            send_amount=send_amount_raw)

    async def blockgen_single_account_opener(
            self,
            representative=None,
            source_key=None,  #
            source_seed=None,
            source_index=None,
            destination_key=None,  #
            destination_seed=None,
            destination_index=None,
            send_amount=1,
            number_of_accounts=1000,
            accounts_keep_track=False,
            increment_index=False):
        if accounts_keep_track:
            if self.single_account_open_counter >= number_of_accounts:
                return []
            if increment_index:
                destination_index = self.single_account_open_counter
        self.single_account_open_counter = self.single_account_open_counter + 1

        destination = self.nano_lib.nanolib_account_data(
            private_key=destination_key,
            seed=destination_seed,
            index=destination_index)
        source = self.nano_lib.nanolib_account_data(private_key=source_key,
                                                    seed=source_seed,
                                                    index=source_index)

        send_block = await self.nano_rpc_default.create_send_block_pkey(
            source["private"],
            destination["account"],
            send_amount,
            broadcast=self.broadcast)

        open_block = await self.nano_rpc_default.create_open_block(
            destination["account"],
            destination["private"],
            send_amount,
            representative,
            send_block["hash"],
            broadcast=self.broadcast)
        open_block["account_data"]["source_seed"] = destination_seed
        open_block["account_data"]["source_index"] = destination_index

        res = [send_block, open_block]
        return res



from nanolab.loggers.factories.logger_factory import LoggerFactory
from nanolab.loggers.factories.sink_factory import SinkFactory
from nanolab.loggers.contracts import ILogger, ISink
from nanolab.src.utils import get_config_parser
import asyncio
from typing import List, Dict


class LoggingManager:

    def __init__(self, logger_params, sink_params: List[Dict]):
        self.nodes_config = get_config_parser().get_nodes_config()
        self.included_peers = logger_params.pop("included_peers", None)
        self.excluded_peers = logger_params.pop("excluded_peers", [])
        self.interval = logger_params.pop("interval", 1)
        self.logger_type = logger_params.pop("type", "rpc")

        self.logger_params = logger_params
        self.sink_params_list = sink_params

    async def create_logger_and_storages(self, current_logger_params):
        logger = await LoggerFactory.create_logger(self.logger_type,
                                                   current_logger_params)

        storages = []
        for sink_params in self.sink_params_list:
            sink_type = sink_params.pop("type", "console")
            storage = SinkFactory.create_storage(sink_type, sink_params)
            storages.append(storage)

        return logger, storages

    async def create_loggers(self):
        for node in self.nodes_config:
            if (self.included_peers and node["name"] not in self.included_peers
                    ) or node["name"] in self.excluded_peers:
                continue
            current_params = dict(self.logger_params)
            current_params["node_name"] = node["name"]
            current_params["rpc_url"] = node["rpc_url"]
            yield await self.create_logger_and_storages(current_params)

    async def start_logging(self):
        tasks = []
        async for logger, storages in self.create_loggers():
            tasks.append(
                self.perform_logging_task(logger, storages, self.interval))
        await asyncio.gather(*tasks)

    async def perform_logging_task(self, logger: ILogger,
                                   storages: List[ISink], interval: int):
        async for logs in logger.fetch_logs():
            for storage in storages:
                storage.store_logs(logs)
            await asyncio.sleep(interval)
        for storage in storages:
            storage.end()



class BlockConfirmationEvent:

    def __init__(self,
                 block_hash: str,
                 confirmed: bool,
                 conf_duration: float = None):
        self.block_hash = block_hash
        self.confirmed = confirmed
        self.conf_duration = conf_duration



import os
import requests
from pathlib import Path
import json


class DataResolver():

    def resolve_config(self, path):

        if path.startswith('http') or path.startswith('https'):
            return requests.get(path).text
        else:
            with open(os.path.abspath(path), 'r') as f:
                return f.read()

    def resolve_config_elements(self, config_path):
        """
        Resolves all the elements from a config file, downloading any remote URLs to disk paths.
        """
        with open(config_path, 'r') as f:
            config = json.load(f)

        for key in config.get("global", {}).keys():
            if isinstance(config["global"][key],
                          str) and config["global"][key].startswith('http'):
                url = config["global"][key]
                filename = Path(url).name
                destination_path = os.path.join(os.path.dirname(config_path),
                                                filename)
                self.download_file(url, destination_path)
                config["global"][key] = destination_path

        return config

    def download_file(self, url: str, destination_path: str):
        """
        Downloads a file from the given URL to the local destination path.
        Skips the download if the file already exists.
        """
        destination_path = Path(destination_path)
        if destination_path.exists():
            return

        with requests.get(url, stream=True) as response:
            response.raise_for_status()
            with open(destination_path, 'wb') as f:
                for chunk in response.iter_content(chunk_size=8192):
                    f.write(chunk)



from abc import ABC, abstractmethod
from nanorpc.client import NanoRpcTyped
from nanolab.publisher.block_event import BlockConfirmationEvent
from nanolab.src.utils import get_config_parser
import asyncio
import time


class IBlockAsserts(ABC):

    @abstractmethod
    async def assert_blocks_published(self, blocks):
        pass

    @abstractmethod
    async def assert_all_blocks_cemented(self):
        pass

    @abstractmethod
    async def assert_blocks_confirmed(self, block_hashes):
        pass

    @abstractmethod
    async def assert_blocks_confirmed_wait(self, block_hashes, wait_s,
                                           interval):
        pass

    @abstractmethod
    async def assert_single_block_confirmed(self, block_hash: str):
        pass

    @abstractmethod
    async def assert_single_block_confirmed_wait(self, block_hash, wait_s,
                                                 interval):
        pass


# This class contains the actual implementation of our block assertions.


class BlockAsserts(IBlockAsserts):

    def __init__(self, event_bus=None, node_name=None):
        self.event_bus = event_bus
        conf_p = get_config_parser()
        node_name = node_name if node_name else conf_p.get_nodes_name()[:-1]
        self.nano_rpc_default = NanoRpcTyped(conf_p.get_node_rpc(node_name))
        self.nano_rpc_all = [NanoRpcTyped(url)
                             for url in conf_p.get_nodes_rpc()]

    async def _fetch_block_info(self, block_hash):
        return await self.nano_rpc_default.block_info(block_hash,
                                                      json_block="true")

    async def _fetch_all_block_info(self, block_hashes, filter_func=None):
        tasks = [
            self._fetch_block_info(block_hash) for block_hash in block_hashes
        ]
        responses = await asyncio.gather(*tasks)

        # The mapping relies on the fact the asyncio.gather function preserves the order of results,
        # meaning that results are returned in the order of the input awaitable objects.
        responses = [{
            **response, 'hash': hash
        } for response, hash in zip(responses, block_hashes)]
        if filter_func:
            responses = list(filter(filter_func, responses))
        return responses

    async def assert_blocks_published(self, blocks):
        block_infos = await self._fetch_all_block_info(blocks)

        assert all(info['exists']
                   for info in block_infos), f"Some blocks were not published"

    async def assert_all_blocks_cemented(self):
        for rpc in self.nano_rpc_all:
            block_count = await rpc.block_count()
            assert block_count["count"] == block_count[
                "cemented"], "Not all blocks are cemented"

    async def _blocks_confirmed(self, block_hashes):
        confirmed_blocks = set()
        for block_hash in block_hashes:
            block_info = await self._fetch_block_info(block_hash)
            if block_info.get('confirmed') == "true":
                confirmed_blocks.add(block_hash)
        return confirmed_blocks

    async def _unconfirmed_blocks(self, block_hashes, batch_size=None):
        batch = list(block_hashes)[:batch_size] if batch_size else block_hashes
        confirmed_blocks = await self._blocks_confirmed(batch)
        unconfirmed_blocks = set(block_hashes) - confirmed_blocks
        return unconfirmed_blocks

    async def assert_blocks_confirmed(self, block_hashes):
        unconfirmed_blocks = await self._unconfirmed_blocks(
            block_hashes, batch_size=len(block_hashes))
        if unconfirmed_blocks:
            raise AssertionError(
                f"{len(list(unconfirmed_blocks))} blocks are not confirmed")

    async def _publish_block_confirmation_event(self, wait_s, block_hash,
                                                timeout):
        if not self.event_bus:
            return
        event = BlockConfirmationEvent(block_hash, timeout, wait_s)
        await self.event_bus.publish('block_confirmed', event)

    async def assert_blocks_confirmed_wait(self, block_hashes, wait_s,
                                           interval):
        batch_size = 1000
        unconfirmed_blocks = set(block_hashes)
        start_time = time.time()
        while time.time() - start_time < wait_s and unconfirmed_blocks:
            confirmed_blocks = await self._blocks_confirmed(
                list(unconfirmed_blocks)[:batch_size])
            for block_hash in confirmed_blocks:
                conf_duration = time.time() - start_time
                await self._publish_block_confirmation_event(
                    conf_duration, block_hash, True)
                unconfirmed_blocks.remove(block_hash)
            if unconfirmed_blocks:
                await asyncio.sleep(interval)

        if unconfirmed_blocks:
            # If there are any remaining unconfirmed blocks, raise an AssertionError and send the events.
            for block_hash in unconfirmed_blocks:
                await self._publish_block_confirmation_event(
                    wait_s, block_hash, False)
            raise AssertionError(
                f"{len(list(unconfirmed_blocks))} unconfirmed blocks! e.g. {list(unconfirmed_blocks)[0]}"
            )

    async def assert_single_block_confirmed(self, block_hash: str):
        await self.assert_blocks_confirmed([block_hash])

    async def assert_single_block_confirmed_wait(self, block_hash, wait_s,
                                                 interval):
        await self.assert_blocks_confirmed_wait([block_hash], wait_s, interval)



from os import environ
import inspect
import logging


class CommandMixinBase:

    def __init__(self, command_instance):
        self.command_instance = command_instance
        self.logger = logging.getLogger(__name__)

    def _get_completed_variables(self):
        variables = self.command_instance.command_config.get('variables', {})

        # add global variables
        variables["docker_tag"] = environ["docker_tag"]
        return variables

    def _get_filtered_variables(self, method):
        variables = self._get_completed_variables()
        method_parameters = self._get_method_parameters(method)
        return {
            key: value
            for key, value in variables.items() if key in method_parameters
        }

    def _get_method_parameters(self, method):
        signature = inspect.signature(method)
        return [param.name for param in signature.parameters.values()]



from nanolab.loggers.sinks.csv_sink import CSVSink
from nanolab.loggers.sinks.sql_sink import SqlSink
from nanolab.loggers.sinks.console_sink import ConsoleSink


class SinkFactory:

    @staticmethod
    def create_storage(sink_type, config_params_dict):
        if sink_type == 'console':
            return ConsoleSink(**config_params_dict)
        elif sink_type == 'sql':
            return SqlSink(**config_params_dict)
        elif sink_type == 'csv':
            return CSVSink(**config_params_dict)
        else:
            raise ValueError(f"Invalid storage_type: {sink_type}")



from nanorpc.client import NanoRpcTyped
from nanolab.src.utils import get_config_parser
from time import strftime, gmtime, time
from math import ceil


class ConfirmationStatsManager:

    def __init__(self, block_timeout_s):
        self.calculator = ConfirmationCalculator(block_timeout_s)
        self.printer = ConfirmationStatsPrinter()
        self.formatter = ConfirmationTableFormatter()

    async def initialize(self, node_name):
        self.nanorpc = NanoRpcTyped(
            get_config_parser().get_node_rpc(node_name))
        start_block_count = await self.nanorpc.block_count()
        self.calculator.set_start_block_count(start_block_count)

    def set_end_block_count(self, end_block_count: dict) -> None:
        self.calculator.set_end_block_count(end_block_count)

    async def print_stats(self) -> None:
        end_block_count = await self.nanorpc.block_count()
        self.calculator.set_end_block_count(end_block_count)
        stats = self.calculator.compute_stats()
        self.printer.print_stats(stats, self.formatter)

    async def on_block_confirmed(self, event):
        # Do something with the event, for example, append it to a list
        self.calculator.add_event(event)


class ConfirmationCalculator:

    def __init__(self, block_timeout_s):
        self.block_timeout_s = block_timeout_s
        self.start_time = time()
        self.events = []  # Add this list to keep track of events

    def _percentile(self, data, percentile):
        n = len(data)
        p = n * percentile / 100
        if p.is_integer():
            return sorted(data)[int(p)]
        else:
            return sorted(data)[int(ceil(p)) - 1]

    def add_event(self, event):
        # You might want to change the structure of the data in the list
        self.events.append({
            "conf_duration": event.conf_duration,
            "timeout": not event.confirmed
        })

    def set_start_block_count(self, start_block_count):
        self.start_block_count = start_block_count

    def set_end_block_count(self, end_block_count):
        self.end_block_count = end_block_count

    def compute_stats(self):
        self.new_blocks = int(self.end_block_count.get("count", 0)) - int(
            self.start_block_count.get("count", 0))
        self.new_cemented = int(self.end_block_count.get("cemented", 0)) - int(
            self.start_block_count.get("cemented", 0))
        self.confirmations = [
            e["conf_duration"] for e in self.events if e["timeout"] == False
        ]
        self.timeouts = [e for e in self.events if e["timeout"]]
        self.conf_duration = time() - self.start_time

        gather_int = {
            "confs":
            len(self.confirmations),
            "timeouts":
            len(self.timeouts),
            "bps":
            self.new_blocks / self.conf_duration,
            "cps":
            self.new_cemented / self.conf_duration,
            "min_conf_s":
            min(self.confirmations) if len(self.confirmations) > 0 else -1,
            "max_conf_s":
            max(self.confirmations) if len(self.confirmations) > 0 else -1,
            "perc_50_s":
            self._percentile(self.confirmations, 50)
            if len(self.confirmations) > 0 else -1,
            "perc_75_s":
            self._percentile(self.confirmations, 75)
            if len(self.confirmations) > 0 else -1,
            "perc_90_s":
            self._percentile(self.confirmations, 90)
            if len(self.confirmations) > 0 else -1,
            "perc_99_s":
            self._percentile(self.confirmations, 99)
            if len(self.confirmations) > 0 else -1,
            "block_timeout_s":
            self.block_timeout_s,
            "total_s":
            self.conf_duration,
            "new_blocks":
            self.new_blocks,
            "new_cemented":
            self.new_cemented,
        }
        return gather_int


class ConfirmationTableFormatter:

    def format_table(self, table, print_header=True):
        col_width = [max(len(str(x)) for x in col) for col in zip(*table)]
        formatted_table = []
        for line in table:
            if not print_header:
                print_header = True
                continue
            formatted_table.append(" | ".join("{:{}}".format(x, col_width[i])
                                              for i, x in enumerate(line)) +
                                   " |")
        return formatted_table


class ConfirmationStatsPrinter:

    def print_stats(self, stats, formatter):
        table_pr1 = (stats.keys(),
                     [str(round(stats[x], 2)).ljust(8) for x in stats])
        formatted_table = formatter.format_table(table_pr1, print_header=True)
        print("\n")
        for line in formatted_table:
            print(f'{strftime("%H:%M:%S", gmtime())} {line}')



from nanolab.xnomin.peers import get_connected_socket_endpoint, message_header, block_state, block_type_enum, message_type_enum, network_id, message_type, get_peers_from_service
from nanolab.xnomin.handshake import node_handshake_id
from nanomock.modules.nl_parse_config import ConfigReadWrite
from nanolab.src.utils import get_config_parser
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any, Dict, List
import asyncio
import random
import time
import itertools
import threading


from nanolab.loggers.logger_manager import LoggingManager


async def start_loggers(logger_params: Dict, sink_params: List[Dict]):
    logger_manager = LoggingManager(logger_params, sink_params)
    await logger_manager.start_logging()


async def xnolib_publish(params: dict):

    block_lists = get_blocks_from_disk(params)
    sp = SocketPublish(params)
    messages, _ = sp.flatten_messages(block_lists)
    sp_task = asyncio.create_task(sp.run(messages))
    await asyncio.gather(sp_task)


def read_blocks_from_disk(path, seeds=False, hashes=False, blocks=False):
    res = ConfigReadWrite().read_json(path)
    if seeds:
        return res["s"]
    if hashes:
        return res["h"]
    if blocks:
        return res["b"]
    return res


def set_default_params(params: dict, all_blocks: dict):
    params.setdefault("start_round", 0)
    params.setdefault("end_round", len(all_blocks["h"]))
    params.setdefault("subset", {})
    params["subset"].setdefault("start_index", 0)
    params["subset"].setdefault(
        "end_index", max(len(inner_list) for inner_list in all_blocks["h"]))

    return params


def get_block_subset(all_blocks: dict, start_round: int, end_round: int,
                     start_index: int, end_index: int):
    blocks = {}
    blocks['b'] = [
        x[start_index:end_index]
        for x in all_blocks['b'][start_round:end_round]
    ]
    blocks['h'] = [
        x[start_index:end_index]
        for x in all_blocks['h'][start_round:end_round]
    ]
    return blocks


def get_blocks_from_disk(params: dict):
    # mandatory params: blocks_path
    # optional params: start_round, end_round, subset
    blocks_path = params["blocks_path"]
    all_blocks = read_blocks_from_disk(blocks_path)

    params = set_default_params(params, all_blocks)
    subset_start_index = int(params["subset"]["start_index"])
    subset_end_index = int(params["subset"]["end_index"])
    start_round = int(params["start_round"])
    end_round = int(params["end_round"])

    blocks = get_block_subset(all_blocks, start_round,
                              end_round, subset_start_index,
                              subset_end_index)

    return blocks


class SocketPublish:

    def __init__(self, params: Dict[str, Any]):
        self.bps = float(params["bps"])
        self.peers = params.get("peers")
        self.split = params.get("split", False)
        # skips 1st socket (genesis)
        self.split_skip = params.get("split_skip", False)
        self.reverse = params.get("reverse", False)
        self.shuffle = params.get("shuffle", False)
        self.sockets, self.hdr = self.__set_sockets_handshake()

    def handshake_peer(self, peeraddr, peerport, ctx):
        try:
            print('Connecting to [%s]:%s' % (peeraddr, peerport))
            s = get_connected_socket_endpoint(peeraddr, peerport)
            signing_key, verifying_key = node_handshake_id.keypair()
            node_handshake_id.perform_handshake_exchange(
                ctx, s, signing_key, verifying_key)
            return s
        except Exception as e:
            print(str(e))
            pass

    class msg_publish:

        def __init__(self, hdr, block):
            assert (isinstance(hdr, message_header))
            self.hdr = hdr
            self.block = block

        def serialise(self):
            data = self.hdr.serialise_header()
            data += self.block.serialise(False)
            return data

        def __str__(self):
            return str(self.hdr) + "\n" + str(self.block)

    def get_xnolib_context(self, peers=None):
        conf_p = get_config_parser()
        ctx = conf_p.get_xnolib_localctx()
        if  conf_p.get_env() == 'beta' :
            ctx["net_id"] = network_id(ord('B'))
        elif  conf_p.get_env() == 'live' :
            ctx["net_id"] = network_id(ord('C'))
        else :
            ctx["net_id"] = network_id(ord('X'))


        if peers is not None:  # chose a single peer , if enabled in config file
            for peer in ctx["peers"].copy():
                if peer not in peers:
                    ctx["peers"].pop(peer, None)
        return ctx

    def __set_sockets_handshake(self):
        ctx = self.get_xnolib_context(peers=self.peers)

        msgtype = message_type_enum.publish
        hdr = message_header(ctx['net_id'], [21, 21, 20],
                            message_type(msgtype), 0)
        hdr.set_block_type(block_type_enum.state)
        all_peers = get_peers_from_service(ctx)
        sockets = []

        def handshake_peer_wrapper(peer):
            s = self.handshake_peer(str(peer.ip), peer.port, ctx)
            if s is not None:
                return {
                    "socket": s,
                    "peer": f"{peer.ip}:{peer.port}"
                }
            return None

        # Adjust max_workers according to your needs
        with ThreadPoolExecutor(max_workers=20) as executor:
            futures = {executor.submit(handshake_peer_wrapper, peer): peer for peer in all_peers}
            for future in as_completed(futures):
                result = future.result()
                if result is not None:
                    sockets.append(result)

        return sockets, hdr

    def flatten_messages(self,
                         block_lists: List[List[Dict[str, Any]]]) -> List[Any]:
        messages = []
        for block_list in block_lists['b']:
            publish_msg = [
                self.msg_publish(self.hdr, block_state.parse_from_json(block))
                for block in block_list
            ]
            messages.extend(publish_msg)

        block_hashes = list(itertools.chain(*block_lists['h']))

        return messages, block_hashes

    async def publish_message(self, socket: Dict[str, Any], messages: List[Any]) -> None:
        if self.bps <= 0:
            raise ValueError("bps must be greater than 0")

        start_time = time.time()
        # Calculate time interval between messages based on bps
        message_interval = 1 / self.bps
        sent_messages = 0
        # accumulated_time is required because asyncio.sleep(accumulated_time) accepts time in miliseconds only
        accumulated_time = 0

        for idx, message in enumerate(messages):
            try:
                # Send the message
                socket['socket'].sendall(message.serialise())
                sent_messages += 1

                # Calculate time to wait before sending the next message
                end_time = time.time()
                elapsed_time = end_time - start_time
                remaining_time = max(message_interval - elapsed_time, 0)
                accumulated_time += remaining_time

                # If accumulated_time reaches or exceeds 1 millisecond, sleep for accumulated_time and reset it
                if accumulated_time >= 0.001:
                    await asyncio.sleep(accumulated_time)
                    accumulated_time = 0  # Reset after sleeping

                # Reset start time for the next message
                start_time = time.time()
            except Exception as e:
                print(
                    f"Error sending message {idx+1} to {socket['peer']}: {str(e)}")

    # async def publish_message(self, socket: Dict[str, Any],
    #                           messages: List[Any]) -> None:
    #     start_time = time.time()
    #     sent_messages = 0
    #     for idx, message in enumerate(messages):
    #         try:
    #             socket['socket'].sendall(message.serialise())
    #             sent_messages += 1
    #             if sent_messages % self.bps == 0:
    #                 end_time = time.time()
    #                 elapsed_time = end_time - start_time
    #                 remaining_time = max(1 - elapsed_time, 0)
    #                 await asyncio.sleep(remaining_time)
    #                 start_time = time.time()
    #         except Exception as e:
    #             print(
    #                 f"Error sending message {idx+1} to {socket['peer']}: {str(e)}"
    #             )

    def create_publish_tasks(self, sockets: List[Dict[str, Any]],
                             messages: List[Any]) -> List[asyncio.Task]:
        if self.split:
            tasks = self.create_split_tasks(sockets, messages)
        elif self.split_skip:
            tasks = self.create_split_tasks(sockets,
                                            messages,
                                            skip_first_socket=True)
        elif self.shuffle:
            tasks = self.create_shuffle_tasks(sockets, messages)
        elif self.reverse:
            tasks = self.create_reverse_tasks(sockets, messages)
        else:
            tasks = self.create_default_tasks(sockets, messages)
        return tasks

    def consume_and_discard(self, socket_info):
        sock = socket_info['socket']
        # peer = socket_info['peer']
        while True:
            if self.read_socket(sock, 1024) is None:
                time.sleep(0.25)
                # print(f"No data from {peer}. Wait 100ms")

    def read_socket(self, sock, byte_count: int) -> bytes or None:
        try:
            data = sock.recv(byte_count)
            if not data:
                return None
            return data
        except OSError as msg:
            # Silently handle timeout errors and other OSError exceptions
            print(f"Error reading from socket (peer: {sock.getpeername()}): {msg}")
            return None


    def create_split_tasks(self, sockets, messages, skip_first_socket=False):
        num_sockets = len(sockets) - 1 if skip_first_socket else len(sockets)
        messages_per_socket, remainder = divmod(len(messages), num_sockets)
        tasks = []
        sockets_to_use = sockets[1:] if skip_first_socket else sockets
        for i, socket in enumerate(sockets_to_use,
                                   start=int(skip_first_socket)):
            start = i * messages_per_socket + min(i, remainder)
            end = start + messages_per_socket + (1 if i < remainder else 0)
            tasks.append(self.publish_message(socket, messages[start:end]))
        return tasks

    def create_shuffle_tasks(self, sockets, messages):
        tasks = [
            self.publish_message(socket, random.sample(messages,
                                                       len(messages)))
            for socket in sockets
        ]
        return tasks

    def create_reverse_tasks(self, sockets, messages):
        messages.reverse()
        tasks = self.create_default_tasks(sockets, messages)
        return tasks

    def create_default_tasks(self, sockets, messages):
        tasks = [self.publish_message(socket, messages) for socket in sockets]
        return tasks

    async def run(self, messages: List[Any]) -> int:
        for socket_info in self.sockets:
            threading.Thread(target=self.consume_and_discard, args=(socket_info,), daemon=True).start()


        tasks = self.create_publish_tasks(self.sockets, messages)
        await asyncio.gather(*tasks)
        # make sure the last few blocks are published.
        # ideally this would check if all messages are received
        await asyncio.sleep(15)

        message_count = len(messages)
        return message_count



from os import environ, path, walk
import shutil
from importlib import resources
from pathlib import Path
import logging
import requests
from nanomock.modules.nl_parse_config import ConfigParser
import filecmp

logger = logging.getLogger(__name__)


def _copy_changed_files(src_file: str, dst_file: str):
    """Copy files from source to destination if they have changed."""
    if not path.exists(dst_file) or not filecmp.cmp(
            src_file, dst_file, shallow=False):
        shutil.copy2(src_file, dst_file)


def _walk_and_copy(src_dir: str, dst_dir: str):
    """Walk through the source directory and copy changed files to destination."""
    for src_root, _, files in walk(src_dir):
        dst_root = src_root.replace(src_dir, dst_dir, 1)
        for file in files:
            src_file = path.join(src_root, file)
            dst_file = path.join(dst_root, file)
            _copy_changed_files(src_file, dst_file)


def extract_packaged_data_to_disk(snippets_path):
    with resources.path("nanolab", "__init__.py") as src_dir:
        src_dir = src_dir.parent / "snippets"
        dest_dir = snippets_path

        if src_dir.exists():
            dest_dir.mkdir(parents=True, exist_ok=True)
            _walk_and_copy(str(src_dir), str(dest_dir))
            logger.info(
                f"nanolab default snippets have been copied into {snippets_path}."
            )
        else:
            raise FileExistsError("nanolab.snippets not found.")


def set_nanomock_config_path(config_path: Path):
    environ["NL_CONF_DIR"] = config_path.parent
    environ["NL_CONF_FILE"] = config_path.name


def get_config_parser() -> ConfigParser:
    default_path = "."
    default_file = "nl_config.toml"
    config_parser = ConfigParser(environ.get("NL_CONF_DIR", default_path),
                                 environ.get("NL_CONF_FILE", default_file))

    return config_parser


def download_data(data: str, filename: str, resource_path: Path) -> Path:
    destination = resource_path / filename.strip()

    if not destination.exists():
        print(f"Start Fetching '{filename}' to '{destination}'")
        if data.startswith("http://") or data.startswith("https://"):
            _download_remote_data(data, destination)
        else:
            _copy_local_data(data, destination)
        print(f"Done  Fetching '{filename}' to '{destination}'")
    else:
        print(f"Skip  Fetching '{filename}' to '{destination}'")

    return destination


def _download_remote_data(data: str, destination: Path) -> None:
    response = requests.get(data)
    if response.status_code == 200:
        with destination.open("wb") as f:
            f.write(response.content)

    else:
        raise requests.HTTPError(
            f"Failed to fetch '{destination.name}' from '{data}', status code: {response.status_code}"
        )


def _copy_local_data(data: str, destination: Path) -> None:
    with destination.open("w") as f:
        f.write(data)


def print_dot(func):

    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        print('.', end='', flush=True)
        return result

    return wrapper


import base64
import hashlib


# this function expects account to be a 32 byte bytearray
def to_account_addr(account: bytes, prefix: str = 'nano_') -> str:
    assert (len(account) == 32)

    RFC_3548 = b"ABCDEFGHIJKLMNOPQRSTUVWXYZ234567"
    ENCODING = b"13456789abcdefghijkmnopqrstuwxyz"

    h = hashlib.blake2b(digest_size=5)
    h.update(account)
    checksum = h.digest()

    # prefix account to make it even length for base32, add checksum in reverse byte order
    account2 = b'\x00\x00\x00' + account + checksum[::-1]

    # use the optimized base32 lib to speed this up
    encode_account = base64.b32encode(account2)

    # simply translate the result from RFC3548 to Nano's encoding, snip off the leading useless bytes
    encode_account = encode_account.translate(
        bytes.maketrans(RFC_3548, ENCODING))[4:]

    # add prefix, label and return
    return prefix + encode_account.decode()


def account_key(account: str) -> bytes:
    """Get the public key for account
    :param str account: account name e.g. nano_31fr1qtbrfnujcspx5xq61uxgjf9j6rzckdj1kdn61y3h53nxr7911dzetk3
    :return: 32 byte public key
    :rtype: bytes
    :raise AssertionError: for invalid account
    """
    account_prefix = "nano_"
    _B32 = b"13456789abcdefghijkmnopqrstuwxyz"
    assert (len(account) == len(account_prefix) + 60
            and account[:len(account_prefix)] == account_prefix)

    account = b"1111" + account[-60:].encode()
    account = account.translate(bytes.maketrans(_B32, base64._b32alphabet))
    key = base64.b32decode(account)

    checksum = key[:-6:-1]
    key = key[3:-5]

    assert hashlib.blake2b(key, digest_size=5).digest() == checksum

    return key


class EventBus:

    def __init__(self):
        self.listeners = {}

    def subscribe(self, event_type, listener):
        if event_type not in self.listeners:
            self.listeners[event_type] = []
        self.listeners[event_type].append(listener)

    async def publish(self, event_type, event):
        if event_type in self.listeners:
            for listener in self.listeners[event_type]:
                await listener(event)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name="nanolab",
      version="0.0.27",
      author="gr0vity",
      description="testing tool using nanomock",
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://github.com/gr0vity-dev/nanolab",
      packages=find_packages(exclude=["unit_tests"]),
      include_package_data=True,
      install_requires=["nanomock>=0.0.14",
                        "py_ed25519_blake2b", "sqlalchemy", "requests", "aiohttp", "nanolog_parser"],
      entry_points={
          'console_scripts': [
              'nanolab=nanolab.main:main',
          ],
      })



from nanolab.xnomin.peers import message_header, py_ed25519_blake2b, socket, hexlify, message_type
from secrets import token_bytes
from typing import Tuple
import time


def read_socket(sock: socket.socket, byte_count: int, timeout: float = 30.0) -> bytes or None:
    try:
        sock.settimeout(timeout)  # Set timeout for socket operations
        data = bytearray()
        start_time = time.time()
        while len(data) < byte_count:
            if time.time() - start_time > timeout:
                raise socket.timeout("Timeout while reading socket")
            chunk = sock.recv(min(byte_count - len(data), 1024))
            if not chunk:
                raise ValueError('SocketClosedByPeer read_socket: data=%s' % data)
            data += chunk
        return bytes(data)
    except (OSError, socket.timeout) as msg:
        print('[read_socket] Error reading %d bytes, data=%s, msg=%s' %
              (byte_count, hexlify(data), msg))
        return None


class node_handshake_id:

    @classmethod
    def parse(cls, hdr: message_header, payload: bytes):
        if hdr.is_query() and hdr.is_response():
            handshake = handshake_response_query.parse_query_response(
                hdr, payload)
        elif hdr.is_query():
            handshake = handshake_query.parse_query(hdr, payload)
        elif hdr.is_response():
            handshake = handshake_response.parse_response(hdr, payload)
        return handshake

    @classmethod
    def keypair(
        cls
    ) -> Tuple[py_ed25519_blake2b.SigningKey, py_ed25519_blake2b.VerifyingKey]:
        return py_ed25519_blake2b.create_keypair()

    @classmethod
    def perform_handshake_exchange(
            cls, ctx: dict, s: socket.socket,
            signing_key: py_ed25519_blake2b.SigningKey,
            verifying_key: py_ed25519_blake2b.VerifyingKey) -> bytes:
        hdr = message_header(ctx['net_id'], [21, 21, 20], message_type(10), 1)
        msg_handshake = handshake_query(hdr)
        s.sendall(msg_handshake.serialise())
        try:
            data = read_socket(s, 136)
            hdr = message_header.parse_header(data[0:8])
            recvd_response = handshake_response_query.parse_query_response(
                hdr, data[8:])

            response = handshake_response.create_response(
                ctx, recvd_response.cookie, signing_key, verifying_key)
            s.sendall(response.serialise())

            # vk = py_ed25519_blake2b.keys.VerifyingKey(recvd_response.account)
            # vk.verify(recvd_response.sig, msg_handshake.cookie)
        except TypeError:
            raise ValueError("HandshakeExchangeFail")

        return recvd_response.account


class handshake_response_query(node_handshake_id):

    def __init__(self, hdr: message_header, cookie: bytes, account: bytes,
                 signature: bytes):
        assert isinstance(hdr, message_header)
        assert hdr.is_query()
        assert hdr.is_response()

        self.header = hdr
        self.cookie = cookie
        self.account = account
        self.sig = signature

    def serialise(self) -> bytes:
        data = self.header.serialise_header()
        data += self.cookie
        data += self.account
        data += self.sig
        return data

    @classmethod
    def parse_query_response(cls, hdr: message_header, data: bytes):
        assert isinstance(hdr, message_header)
        assert (len(data) == 128)

        cookie = data[0:32]
        account = data[32:64]
        sig = data[64:]

        assert len(sig) == 64

        return handshake_response_query(hdr, cookie, account, sig)

    @classmethod
    def create_response(cls, ctx, cookie,
                        signing_key: py_ed25519_blake2b.SigningKey,
                        verifying_key: py_ed25519_blake2b.VerifyingKey):
        my_cookie = token_bytes(32)
        sig = signing_key.sign(cookie)
        hdr = message_header(ctx['net_id'], [21, 21, 20], message_type(10), 3)
        return handshake_response_query(hdr, my_cookie,
                                        verifying_key.to_bytes(), sig)

    def __str__(self):
        string = "Header: [%s]\n" % str(self.header)
        string += "Cookie: %s\n" % hexlify(self.cookie)
        string += "Account: %s\n" % hexlify(self.account)
        string += "Signature: %s\n" % hexlify(self.sig)
        string += "Is query: True\n"
        string += "Is response: True\n"
        return string

    def __eq__(self, other):
        if not isinstance(other, handshake_response_query):
            return False
        elif not self.header == other.header:
            return False
        elif not self.cookie == other.cookie:
            return False
        elif not self.account == other.account:
            return False
        elif not self.sig == other.sig:
            return False
        return True


class handshake_query(node_handshake_id):

    def __init__(self, hdr: message_header, cookie: bytes = token_bytes(32)):
        assert isinstance(hdr, message_header)
        assert hdr.is_query()
        self.header = hdr
        self.cookie = cookie

    def serialise(self) -> bytes:
        data = self.header.serialise_header()
        data += self.cookie
        return data

    @classmethod
    def parse_query(cls, hdr: message_header, data: bytes):
        assert (len(data) == 32)
        cookie = data
        return handshake_query(hdr, cookie)

    def __str__(self):
        string = "Header: [%s]\n" % str(self.header)
        string += "Cookie: %s\n" % hexlify(self.cookie)
        string += "Is query: True\n"
        string += "Is response: False\n"
        return string

    def __eq__(self, other):
        if not isinstance(other, handshake_query):
            return False
        elif not self.header == other.header:
            return False
        elif not self.cookie == other.cookie:
            return False
        return True


class handshake_response(node_handshake_id):

    def __init__(self, hdr: message_header, account: bytes, signature: bytes):
        assert isinstance(hdr, message_header)
        assert hdr.is_response()

        self.header = hdr
        self.account = account
        self.sig = signature

    def serialise(self) -> bytes:
        data = self.header.serialise_header()
        data += self.account
        data += self.sig
        return data

    @classmethod
    def create_response(cls, ctx, cookie,
                        signing_key: py_ed25519_blake2b.SigningKey,
                        verifying_key: py_ed25519_blake2b.VerifyingKey):
        sig = signing_key.sign(cookie)
        hdr = message_header(ctx['net_id'], [21, 21, 20], message_type(10), 2)
        return handshake_response(hdr, verifying_key.to_bytes(), sig)

    @classmethod
    def parse_response(cls, hdr: message_header, data: bytes):
        assert len(data) == 96
        assert isinstance(hdr, message_header)

        account = data[0:32]
        sig = data[32:]

        assert (len(sig) == 64)

        return handshake_response(hdr, account, sig)

    def __str__(self):
        string = "Header: [%s]\n" % str(self.header)
        string += "Account: %s\n" % hexlify(self.account)
        string += "Signature: %s\n" % hexlify(self.sig)
        string += "Is query: False\n"
        string += "Is response: True\n"
        return string

    def __eq__(self, other):
        if not isinstance(other, handshake_response):
            return False
        elif not self.header == other.header:
            return False
        elif not self.account == other.account:
            return False
        elif not self.sig == other.sig:
            return False
        return True



import os
import subprocess
from nanomock.modules.nl_parse_config import ConfigReadWrite
from .snippet_manager import SnippetManager
from nanolab.command.command import Command
from copy import deepcopy


class ConfigLoader:

    def __init__(self, snippet_manager: SnippetManager):
        self.snippet_manager = snippet_manager
        self.conf_rw = ConfigReadWrite()
        self.default_command_executor = "NodeInteraction"

    def read_config(self, config_path: str):
        self.conf_rw.read_json(config_path)

    def load_config(self, config: dict):
        config = self._resolve_snippets(config, [])
        config = self._apply_globals(config)
        config = self._complete_config(config)

        return config

    def _apply_globals(self, config):
        global_variables = config.get("global", {})

        for command in config["commands"]:
            self._apply_variables_to_command(global_variables, command)

        return config

    def _set_default_class(self, command):
        if command.get("type") in ["python"]:
            command.setdefault("class", self.default_command_executor)

    def _complete_config(self, config):
        for command in config["commands"]:
            self._set_default_class(command)

            if command.get("type") == "threaded":
                for python_command in command["commands"]:
                    self._set_default_class(python_command)

        return config

    def _resolve_snippets(self, config, breadcrumb=[]):
        config["commands"] = self._resolve_commands(config["commands"],
                                                    breadcrumb)
        return config

    def _resolve_commands(self, commands, breadcrumb):
        resolved_commands = []
        for command in commands:
            if command.get("type") == "snippet":
                resolved_commands.extend(
                    self._resolve_snippet_command(command, breadcrumb))
            else:
                resolved_commands.append(command)
        return resolved_commands

    def _resolve_snippet_command(self, command, breadcrumb):
        key = command["key"]
        if key in breadcrumb:
            raise ValueError(f"Circular snippet reference detected: {key}")

        snippet = deepcopy(self.snippet_manager.get_snippet_by_key(key))
        variables = command.get("variables", {})
        self._check_mandatory_vars(snippet, variables, key)
        resolved_snippet = self._resolve_snippets(snippet, breadcrumb + [key])
        return self._replace_vars_in_commands(resolved_snippet["commands"],
                                              variables)

    def _check_mandatory_vars(self, snippet, variables, key):
        mandatory_vars = snippet.get("mandatory_vars", [])
        for var in mandatory_vars:
            if var not in variables:
                raise KeyError(
                    f"Mandatory variable '{var}' is missing in '{key}'. Variables defined: {list(variables.keys())}"
                )

    def _replace_vars_in_commands(self, commands, variables):
        for command in commands:
            self._replace_vars_in_dict(command, variables)
        return commands

    def _replace_vars_in_dict(self, item, variables):
        if isinstance(item, dict):
            for key, value in item.items():
                if isinstance(value, (dict, list)):
                    self._replace_vars_in_dict(value, variables)
                else:
                    item[key] = self._replace_vars_in_item(value, variables)
        elif isinstance(item, list):
            for i in range(len(item)):
                if isinstance(item[i], (dict, list)):
                    self._replace_vars_in_dict(item[i], variables)
                else:
                    item[i] = self._replace_vars_in_item(item[i], variables)

    def _replace_vars_in_item(self, item, variables):
        if isinstance(item, str):
            for var, var_value in variables.items():
                placeholder = f'{{{var}}}'
                if item == placeholder:  # if the item is exactly equal to the placeholder
                    return var_value  # return the original variable value
                if placeholder in item:  # if the placeholder is part of the string
                    item = item.replace(
                        placeholder, str(var_value)
                    )  # replace it with string representation of variable
        return item

    def _apply_variables_to_command(self, global_variables: dict,
                                    command: dict):
        if isinstance(command, dict):
            for key, value in command.items():
                if isinstance(value, str):
                    command[key] = self._replace_global_variables(
                        global_variables, value)
                else:
                    self._apply_variables_to_command(global_variables, value)
        elif isinstance(command, list):
            for item in command:
                self._apply_variables_to_command(global_variables, item)

    def _replace_global_variables(self, global_variables: dict, value: str) -> str:
        for global_key, global_value in global_variables.items():
            value = value.replace(f"{{{global_key}}}", str(global_value))
        return value


class ConfigValidator:

    @staticmethod
    def validate_config(config):
        if "commands" not in config:
            raise ValueError(
                "The 'commands' key is missing in the config file.")

        for idx, command in enumerate(config["commands"]):
            command_type = command.get("type")
            if command_type not in ["bash", "snippet", "python", "threaded"]:
                raise ValueError(
                    f"Invalid command type '{command_type}' at index {idx} in 'commands'"
                )


class ConfigCommandExecutor:

    @staticmethod
    def docker_tag_exists(tag):
        result = subprocess.run(['docker', 'image', 'inspect', tag],
                                stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        return result.returncode == 0

    @staticmethod
    def pull_docker_tag(tag):
        if ConfigCommandExecutor.docker_tag_exists(tag):
            print(f"Docker tag '{tag}' already exists locally.")
            return True
        try:
            subprocess.run(['docker', 'pull', tag], check=True)
            return True
        except subprocess.CalledProcessError:
            return False

    @staticmethod
    def execute_commands(config):
        for docker_tag in config["docker_tags"]:
            if ConfigCommandExecutor.pull_docker_tag(docker_tag):
                os.environ["docker_tag"] = docker_tag
                for command_config in config["commands"]:
                    ConfigCommandExecutor.execute_command(command_config)
            else:
                print(
                    f"Error: Unable to pull Docker tag '{docker_tag}'. Skipping commands for this tag.")

    @staticmethod
    def execute_command(command_config):
        command = Command(command_config)
        command.validate()
        command.execute()



import socket
import time
from typing import Optional, Union, List
import ipaddress
import logging
import binascii
from hashlib import blake2b
import py_ed25519_blake2b
import json
import requests
from nanolab.xnomin.acctools import to_account_addr, account_key


def get_peers_from_service(ctx: dict, url=None):
    if url is None:
        url = ctx['peerserviceurl']
    if url == '':
        json_resp = ctx['peers']
    else:
        session = requests.Session()
        resp = session.get(url, timeout=15)
        json_resp = resp.json()
    return [Peer.from_json(r) for r in json_resp.values()]


def get_connected_socket_endpoint(addr: str,
                                  port: int,
                                  bind_endpoint: tuple = None
                                  ) -> socket.socket:
    s = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
    s.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
    s.settimeout(3)
    if bind_endpoint:
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind(bind_endpoint)
    s.connect((addr, port))
    return s


DEFAULT_LOGGER_NAME = ""  # an empty string will use the root logger


def get_logger(logger_name: str = DEFAULT_LOGGER_NAME) -> logging.Logger:
    logger = logging.getLogger(logger_name)
    if not logger.hasHandlers():
        logger.addHandler(logging.NullHandler())

    return logger


logger = get_logger()


def hexlify(data) -> str:
    if data is None:
        return 'None'
    return binascii.hexlify(data).decode("utf-8").upper()


class block_type_enum:
    invalid = 0
    not_a_block = 1
    send = 2
    receive = 3
    open = 4
    change = 5
    state = 6


class message_type_enum:
    invalid = 0x0
    not_a_block = 0x1
    keepalive = 0x2
    publish = 0x3
    confirm_req = 0x4
    confirm_ack = 0x5
    bulk_pull = 0x6
    bulk_push = 0x7
    frontier_req = 0x8
    # deleted 0x9
    node_id_handshake = 0x0a
    bulk_pull_account = 0x0b
    telemetry_req = 0x0c
    telemetry_ack = 0x0d


def message_type_enum_to_str(msg_type: int):
    return next(name for name, value in vars(message_type_enum).items()
                if value == msg_type)


def block_length_by_type(blktype: int) -> int:
    lengths = {2: 152, 3: 136, 4: 168, 5: 136, 6: 216}
    return lengths[blktype]


def node_id_handshake_size(is_query: bool, is_response: bool) -> int:
    size = 0
    if is_query:
        size += 32
    if is_response:
        size += 32 + 64
    return size


class network_id:

    def __init__(self, rawbyte: int):
        self.parse_header(int(rawbyte))

    def parse_header(self, rawbyte: bytes) -> None:
        if not (rawbyte in [ord('X'), ord('B'), ord('C'), ord('L')]):
            raise ValueError("ParseErrorBadNetworkId")
        self.id = rawbyte

    def __str__(self):
        return chr(self.id)

    def __eq__(self, other):
        if not isinstance(other, network_id):
            return False
        elif self.id != other.id:
            return False
        return True


class message_type:

    def __init__(self, num: int) -> None:
        if not (num in range(0, 14)):
            raise ValueError("ParseErrorBadMessageType")
        self.type = num

    def __str__(self):
        return '%s(%s)' % (str(self.type), message_type_enum_to_str(self.type))

    def __eq__(self, other):
        if not isinstance(other, message_type):
            return False
        return self.type == other.type


class message_header:

    def __init__(self, net_id: network_id, versions: List[int],
                 msg_type: message_type, ext: int):
        self.ext = ext
        self.net_id = net_id
        self.ver_max = versions[0]
        self.ver_using = versions[1]
        self.ver_min = versions[2]
        self.msg_type = msg_type
        assert isinstance(self.msg_type, message_type)

    def serialise_header(self) -> bytes:
        header = b""
        header += ord('R').to_bytes(1, "big")
        header += ord(str(self.net_id)).to_bytes(1, "big")
        header += self.ver_max.to_bytes(1, "big")
        header += self.ver_using.to_bytes(1, "big")
        header += self.ver_min.to_bytes(1, "big")
        header += self.msg_type.type.to_bytes(1, "big")
        header += self.ext.to_bytes(2, "little")
        return header

    def is_query(self) -> bool:
        return self.ext & 1

    def is_response(self) -> bool:
        return self.ext & 2

    def set_is_query(self, bool: bool) -> None:
        QUERY_MASK = 0x0001
        self.ext = self.ext & 0xfffe
        if bool:
            self.ext = self.ext | QUERY_MASK

    def set_is_response(self, bool: bool) -> None:
        RESPONSE_MASK = 0x0002
        self.ext = self.ext & 0xfffd
        if bool:
            self.ext = self.ext | RESPONSE_MASK

    def count_get(self) -> int:
        COUNT_MASK = 0xf000
        return (self.ext & COUNT_MASK) >> 12

    def block_type(self) -> int:
        BLOCK_TYPE_MASK = 0x0f00
        return (self.ext & BLOCK_TYPE_MASK) >> 8

    def set_block_type(self, block_type: int) -> None:
        assert (isinstance(block_type, int))
        block_type = block_type << 8
        self.ext = self.ext & 0xf0ff
        self.ext = self.ext | block_type

    def set_item_count(self, count: int) -> None:
        assert (isinstance(count, int))
        count = count << 12
        self.ext = self.ext & 0x0fff
        self.ext = self.ext | count

    @classmethod
    def parse_header(cls, data: bytes):
        assert (len(data) == 8)
        if data[0] != ord('R'):
            raise ValueError("ParseErrorBadMagicNumber")
        net_id = network_id(data[1])
        versions = [data[2], data[3], data[4]]
        msg_type = message_type(data[5])
        ext = int.from_bytes(data[6:], "little")
        return message_header(net_id, versions, msg_type, ext)

    def telemetry_ack_size(self) -> int:
        telemetry_size_mask = 0x3ff
        return self.ext & telemetry_size_mask

    @classmethod
    def from_json(self, json_hdr: dict):
        return message_header(
            network_id(json_hdr['net_id']),
            [json_hdr['ver_max'], json_hdr['ver_using'], json_hdr["ver_min"]],
            message_type(json_hdr['msg_type']), json_hdr['ext'])

    def payload_length_bytes(self) -> Optional[int]:
        if self.msg_type == message_type(message_type_enum.bulk_pull):
            return None

        if self.msg_type == message_type(message_type_enum.bulk_push):
            return 0

        elif self.msg_type == message_type(message_type_enum.telemetry_req):
            return 0

        elif self.msg_type == message_type(message_type_enum.frontier_req):
            return 32 + 4 + 4

        elif self.msg_type == message_type(
                message_type_enum.bulk_pull_account):
            return 32 + 16 + 1

        elif self.msg_type == message_type(message_type_enum.keepalive):
            return 8 * (16 + 2)

        elif self.msg_type == message_type(message_type_enum.publish):
            return block_length_by_type(self.block_type())

        # elif self.msg_type == message_type(message_type_enum.confirm_ack):
        #     return confirm_ack_size(self.block_type(), self.count_get())

        # elif self.msg_type == message_type(message_type_enum.confirm_req):
        #     return confirm_req_size(self.block_type(), self.count_get())

        elif self.msg_type == message_type(
                message_type_enum.node_id_handshake):
            return node_id_handshake_size(self.is_query(), self.is_response())

        elif self.msg_type == message_type(message_type_enum.telemetry_ack):
            return self.telemetry_ack_size()

        else:
            logger.debug(f"Unknown message type: {self.msg_type}")
            return None

    def __eq__(self, other):
        if str(self) == str(other):
            return True

    def __str__(self):
        str = "NetID: %s, " % self.net_id
        str += "VerMaxUsingMin: %s/%s/%s, " % (self.ver_max, self.ver_using,
                                               self.ver_min)
        str += "MsgType: %s, " % self.msg_type
        str += "Extensions: %s" % hexlify(self.ext.to_bytes(2, "big"))
        return str


class block_state:

    def __init__(self, account: bytes, prev: bytes, rep: bytes, bal: int,
                 link: bytes, sig: bytes, work: int):
        assert (isinstance(work, int))
        self.account = account
        self.previous = prev
        self.representative = rep
        self.balance = bal
        self.link = link
        self.signature = sig
        self.work = work
        self.ancillary = {
            "next": None,
            "peers": set(),
            "type": block_type_enum.not_a_block
        }

    def get_previous(self) -> bytes:
        return self.previous

    def get_next(self) -> bytes:
        return self.ancillary['next']

    def get_account(self) -> bytes:
        return self.account

    def get_balance(self) -> int:
        return self.balance

    def get_type_int(self) -> int:
        return block_type_enum.state

    def root(self) -> bytes:
        if int.from_bytes(self.previous, "big") == 0:
            return self.account
        else:
            return self.previous

    def set_type(self, block_type: int) -> None:
        assert block_type in range(block_type_enum.invalid,
                                   block_type_enum.state + 1)
        self.ancillary["type"] = block_type

    def hash_string(self):
        return hexlify(self.hash())

    def hash(self) -> bytes:
        STATE_BLOCK_HEADER_BYTES = (b'\x00' * 31) + b'\x06'
        data = b"".join([
            STATE_BLOCK_HEADER_BYTES, self.account, self.previous,
            self.representative,
            self.balance.to_bytes(16, "big"), self.link
        ])
        return blake2b(data, digest_size=32).digest()

    def serialise(self, include_block_type: bool) -> bytes:
        data = b''
        if include_block_type:
            data += (6).to_bytes(1, "big")
        data += self.account
        data += self.previous
        data += self.representative
        data += self.balance.to_bytes(16, "big")
        data += self.link
        data += self.signature

        # Block states proof of work is received and sent in big endian
        data += self.work.to_bytes(8, "big")
        return data

    def is_epoch_v2_block(self) -> bool:
        if self.link[0:14] == b'epoch v2 block':
            return True
        return False

    def is_epoch_v1_block(self) -> bool:
        if self.link[0:14] == b'epoch v1 block':
            return True
        return False

    def sign(self, signing_key: py_ed25519_blake2b.keys.SigningKey) -> None:
        self.signature = signing_key.sign(self.hash())

    def generate_work(self, min_difficulty: int) -> None:
        root_int = int.from_bytes(self.root(), byteorder='big')
        self.work = pow.find_pow_for_root_and_difficulty(
            root_int, min_difficulty)

    @classmethod
    def parse(cls, data: bytes):
        assert (len(data) == block_length_by_type(6))
        account = data[0:32]
        prev = data[32:64]
        rep = data[64:96]
        bal = int.from_bytes(data[96:112], "big")
        link = data[112:144]
        sig = data[144:208]
        # Block states proof of work is received and sent in big endian
        work = int.from_bytes(data[208:], "big")
        return block_state(account, prev, rep, bal, link, sig, work)

    @classmethod
    def parse_from_json(cls, json_obj: dict):
        assert (json_obj['type'] == 'state')
        account = account_key(json_obj['account'])
        prev = binascii.unhexlify(json_obj['previous'])
        rep = account_key(json_obj['representative'])
        bal = int(json_obj['balance'])
        if len(json_obj['link']) == 64:
            link = binascii.unhexlify(json_obj['link'])
        else:
            link = account_key(json_obj['link'])
        sig = binascii.unhexlify(json_obj['signature'])
        work = int.from_bytes(binascii.unhexlify(json_obj['work']), "big")
        return block_state(account, prev, rep, bal, link, sig, work)

    def to_json(self) -> str:
        jsonblk = {
            'type': 'state',
            'account': to_account_addr(self.account),
            'previous': hexlify(self.previous),
            'representative': to_account_addr(self.representative),
            'balance': str(self.balance),
            'link': hexlify(self.link),
            'link_as_account': to_account_addr(self.link),
            'signature': hexlify(self.signature),
            'work': hexlify(self.work.to_bytes(8, "big"))
        }
        return json.dumps(jsonblk, indent=4)

    def link_to_string(self) -> str:
        if self.link.startswith(b'epoch'):
            return self.link.decode('ascii').replace('\x00', '')
        else:
            return hexlify(self.link)

    # def __str__(self):
    #     hexacc = binascii.hexlify(self.account).decode("utf-8").upper()
    #     string = "------------- Block State -------------\n"
    #     string += "Hash : %s\n" % hexlify(self.hash())
    #     string += "Acc  : %s\n" % hexacc
    #     string += "       %s\n" % to_account_addr(self.account)
    #     string += "Prev : %s\n" % hexlify(self.previous)
    #     string += "Repr : %s\n" % hexlify(self.representative)
    #     string += "       %s\n" % to_account_addr(self.representative)
    #     string += "Bal  : %s\n" % (self.balance / (10**30))
    #     string += "Link : %s\n" % self.link_to_string()
    #     string += "Sign : %s\n" % hexlify(self.signature)
    #     string += "Work : %s\n" % hexlify(self.work.to_bytes(8, "big"))
    #     string += "Next : %s\n" % hexlify(self.ancillary["next"])
    #     string += "Peers: %s" % self.ancillary['peers']
    #     return string

    # def __hash__(self):
    #     return hash((self.account, self.previous, self.representative,
    #                  self.balance, self.link))

    # def __eq__(self, other):
    #     if not isinstance(other, block_state):
    #         return False
    #     elif self.account != other.account:
    #         assert (self.hash() != other.hash())
    #         return False
    #     elif self.previous != other.previous:
    #         assert (self.hash() != other.hash())
    #         return False
    #     elif self.representative != other.representative:
    #         assert (self.hash() != other.hash())
    #         return False
    #     elif self.balance != other.balance:
    #         assert (self.hash() != other.hash())
    #         return False
    #     elif self.link != other.link:
    #         assert (self.hash() != other.hash())
    #         return False
    #     elif self.signature != other.signature:
    #         assert (self.hash() != other.hash())
    #         return False
    #     elif self.work != other.work:
    #         assert (self.hash() != other.hash())
    #         return False
    #     elif self.ancillary != other.ancillary:
    #         assert (self.hash() != other.hash())
    #         return False

    #     return True


class ip_addr:

    def __init__(
            self,
            ipv6: Union[str, ipaddress.IPv6Address] = ipaddress.IPv6Address(0)):
        if isinstance(ipv6, str):
            self.ipv6 = ipaddress.IPv6Address(ipv6)
        else:
            self.ipv6 = ipv6
        assert isinstance(self.ipv6, ipaddress.IPv6Address)

    @classmethod
    def from_string(cls, ipstr: str):
        assert isinstance(ipstr, str)
        a = ipaddress.ip_address(ipstr)
        if a.version == 4:
            ipstr = '::ffff:' + str(a)
        ipv6 = ipaddress.IPv6Address(ipstr)
        return ip_addr(ipv6)

    def serialise(self) -> bytes:
        return self.ipv6.packed

    def is_ipv4(self) -> bool:
        return self.ipv6.ipv4_mapped is not None

    def __str__(self):
        if self.ipv6.ipv4_mapped:
            return '::ffff:' + str(self.ipv6.ipv4_mapped)
        return str(self.ipv6)

    def __eq__(self, other):
        if not isinstance(other, ip_addr):
            return False
        return self.ipv6 == other.ipv6

    def __hash__(self):
        return hash(self.ipv6)


class Peer:

    def __init__(self,
                 ip: ip_addr = ip_addr(),
                 port: int = 0,
                 score: int = -1,
                 is_voting: bool = False,
                 last_seen: int = int(time.time()),
                 incoming: bool = False):
        assert isinstance(ip, ip_addr)
        self.ip = ip
        self.port = port
        self.peer_id = None
        self.is_voting = is_voting
        self.telemetry = None
        self.aux = {}
        self.last_seen = last_seen
        self.incoming = incoming

        # sideband info, not used for equality and hashing
        self.score = score

    def serialise(self) -> bytes:
        data = b""
        data += self.ip.serialise()
        data += self.port.to_bytes(2, "little")
        return data

    def deduct_score(self, score: int) -> None:
        self.score = max(0, self.score - score)

    @classmethod
    def from_json(self, json_peer):
        from nanolab.xnomin.telemetry_req import telemetry_ack
        # Add 'incoming' argument when peer service code gets updated
        peer = Peer(ip_addr(json_peer['ip']), json_peer['port'],
                    json_peer['score'], json_peer['is_voting'])
        if 'telemetry' in json_peer and json_peer['telemetry'] is not None:
            peer.telemetry = telemetry_ack.from_json(json_peer['telemetry'])
        if 'peer_id' in json_peer and json_peer['peer_id']:
            peer.peer_id = binascii.unhexlify(json_peer['peer_id'])
        return peer

    def __str__(self):
        sw_ver = ''
        if self.telemetry:
            sw_ver = ' v' + self.telemetry.get_sw_version()
        return '%s:%s (score:%s, is_voting: %s%s)' % (str(
            self.ip), self.port, self.score, self.is_voting, sw_ver)

    def __eq__(self, other):
        return (self.peer_id is not None
                and self.peer_id == other.peer_id) or (self.ip == other.ip and
                                                       self.port == other.port)

    def __hash__(self):
        return hash((self.ip, self.port))



# decorators.py
import asyncio
import time
from functools import wraps


def ensure_duration(duration=2):

    def decorator(func):

        @wraps(func)
        async def wrapper(*args, **kwargs):
            start_time = time.perf_counter()
            result = await func(*args, **kwargs)
            end_time = time.perf_counter()

            elapsed_time = end_time - start_time
            remaining_time = duration - elapsed_time

            if remaining_time > 0:
                await asyncio.sleep(remaining_time)

            return result

        return wrapper

    return decorator


def print_dot(func):

    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        print('.', end='', flush=True)
        return result

    return wrapper



from .base_mixin import CommandMixinBase
from nanolab import pycmd


class PythonCommandMixin(CommandMixinBase):

    def validate(self):

        if "method" not in self.command_instance.command_config:
            raise ValueError("Python command: 'method' is required.")

        method_name = self.command_instance.command_config["method"]

        if "class" in self.command_instance.command_config:
            cls = getattr(pycmd, self.command_instance.command_config["class"])
            if not cls:
                raise ValueError(
                    f"Python command: Class '{self.command_instance.command_config['class']}' not found."
                )
            if not hasattr(cls, method_name):
                raise ValueError(
                    f"Python command: Method '{method_name}' not found in class '{self.command_instance.command_config['class']}'."
                )
        else:
            if method_name not in globals():
                raise ValueError(
                    f"Python command: Method '{method_name}' not found in the current module."
                )

    def execute(self):
        class_name = self.command_instance.command_config.get(
            "class", "NodeInteraction")
        cls = getattr(pycmd, class_name)
        instance = cls(**self.command_instance.command_config.get(
            'constructor_params', {}))

        method = getattr(instance,
                         self.command_instance.command_config['method'])
        filtered_variables = self._get_filtered_variables(method)
        method(**filtered_variables)



import time
import nanolab.node_interaction as nni
import asyncio
from nanolab.publisher.test_case import TestCaseFactory


class TestClass:

    def long_running_method(self, duration=1):
        time.sleep(duration)

    def test_list(self, test_list):
        if isinstance(test_list, list):
            print(test_list, len(test_list))

    def print_to_console(self, value):
        print(value)

    def return_value(self, value):
        return value


class NodeInteraction:

    def __init__(self):
        # Your existing code...
        self.logger = None

    def start_logger(self, logger_params, sink_params):
        asyncio.run(nni.start_loggers(logger_params, sink_params))

    def publish_create(self, params):
        # Create the test case using the factory
        test_case = TestCaseFactory.create(params)
        asyncio.run(test_case.run())

    def publish_blocks(self, publish_params):
        '''
        mandatory publish_params: blocks_path, bps
         optional publish_params: peers, split, split_skip, reverse, shuffle, 
                  start_round, end_round, subset.start_index, subset.end_index
         '''

        asyncio.run(nni.xnolib_publish(publish_params))



from nanolab.loggers.contracts import ISink, LogData
from nanolab.decorators import print_dot
import csv


class CSVSink(ISink):
    """A storage that writes logs to a CSV file."""

    def __init__(self, **kwargs):
        self.csv_file = kwargs['csv_file']
        self.writer = csv.writer(open(self.csv_file, 'w', newline=''))
        self.writer.writerow([
            "timestamp", "elapsed_time", "check_count", "cemented_count",
            "bps", "cps", "percent_cemented", "percent_checked"
        ])

    @print_dot
    def store_logs(self, logs: LogData):
        """Store logs by writing them to the CSV file."""
        data = [
            logs.timestamp, logs.elapsed_time, logs.check_count,
            logs.cemented_count, logs.bps, logs.cps, logs.percent_cemented,
            logs.percent_checked
        ]

        try:
            self.writer.writerow(data)
        except Exception as e:
            print(f"Failed to write to CSV file: {e}")

    def end(self):
        pass



import requests
import base64


class GitHubAPI:

    def __init__(self, github_user, github_repo, testcase_path, gh_ref = None):
        self.github_user = github_user
        self.github_repo = github_repo
        self.gh_ref = "?ref=" + gh_ref if gh_ref else ""
        self.testcase_path = testcase_path

    def get_api_url(self, endpoint, query="contents/"):
        url = f"https://api.github.com/repos/{self.github_user}/{self.github_repo}/{query}{endpoint}{self.gh_ref}"
        print(url)
        return url

    def extract_ref_from_path(self, testcase_path):
        if "?ref=" in testcase_path:
            parts = testcase_path.split("?ref=")

            # Return testcase_path (without ref) and the ref value
            return parts[0], "?ref=" + parts[1]
        else:
            # Return the original testcase_path and an empty string for gh_ref
            return testcase_path, ""
    
    def get_paths(self):
        response = requests.get(self.get_api_url(""))
        if response.status_code == 200:
            contents = response.json()
            return [item['path'] for item in contents]
        else:
            print(
                "Error retrieving the repository branches. Please check the provided URL.")
            return None
        
    def get_branches(self):
        response = requests.get(self.get_api_url("", "branches"))
        if response.status_code == 200:
            contents = response.json()
            return [item['name'] for item in contents]
        else:
            print(
                "Error retrieving the repository branches. Please check the provided URL.")
            return None    

    def get_files(self):
        response = requests.get(self.get_api_url(self.testcase_path))
        if response.status_code == 200:
            return response.json()
        else:
            print(
                "Error retrieving the repository contents. Please check the provided URL.")
            return None

    def get_file_content(self, file_name):
        file_api_url = self.get_api_url(f"{self.testcase_path}/{file_name}")
        response = requests.get(file_api_url)

        if response.status_code == 200:
            file_content = response.json()["content"]
            return base64.b64decode(file_content).decode("utf-8")
        else:
            print(
                "Error retrieving the file. Please check the provided URL and file name.")
            return None
